# -*- coding: utf-8 -*-
from collections import OrderedDict
from pkg_resources import parse_version
from plone.versioncheck.pypi import check
from plone.versioncheck.pypi import mmbp_tuple
from plone.versioncheck.pypi import PYPI_URL
from plone.versioncheck.pypi import Release
from plone.versioncheck.utils import requests_session

import datetime
import responses


def test_mmbp_tuple():
    assert mmbp_tuple(parse_version(u'0.0.0.0')) == [0, 0, 0, 0]
    assert mmbp_tuple(parse_version(u'1.0')) == [1, 0, 0, 0]
    assert mmbp_tuple(parse_version(u'1.1.0')) == [1, 1, 0, 0]
    assert mmbp_tuple(parse_version(u'0.1')) == [0, 1, 0, 0]
    assert mmbp_tuple(parse_version(u'0.10.1')) == [0, 10, 1, 0]
    assert mmbp_tuple(parse_version(u'1.1.1.1')) == [1, 1, 1, 1]
    assert mmbp_tuple(parse_version(u'1.1.0.a1')) == [1, 1, 0, 0]


demo_json = '''{
    "info": {},
    "releases": {
        "1.0": {},
        "1.0.1": {},
        "1.0.2": {},
        "1.0.3": {},
        "1.0.4": {},
        "1.0.5": {},
        "1.0.6": {},
        "1.0.7": {},
        "1.0.8": {},
        "1.0.9": {},
        "1.0.10": {},
        "1.0.11": {},
        "1.0.12": {},
        "1.0.13.dev0": {},
        "1.1.0": {},
        "1.1.1": {},
        "1.2.0.b1": {},
        "2.0.0": {},
        "3.0.a1": {}
    }
}
'''


assumed_demo_result = OrderedDict([
    ('major', Release(version=u'2.0.0',
                      release_date=datetime.date(1970, 1, 1))),
    ('minor', Release(version=u'1.1.1',
                      release_date=datetime.date(1970, 1, 1))),
    ('bugfix', Release(version=u'1.0.12',
                       release_date=datetime.date(1970, 1, 1))),
    ('majorpre', Release(version=u'3.0.a1',
                         release_date=datetime.date(1970, 1, 1))),
    ('minorpre', Release(version=u'1.2.0.b1',
                         release_date=datetime.date(1970, 1, 1))),
    ('bugfixpre', Release(version=u'1.0.13.dev0',
                          release_date=datetime.date(1970, 1, 1))),
])


@responses.activate
def test_check():
    session = requests_session(nocache=False)
    name = u'demo'
    responses.add(
        responses.GET, '{0}/{1}/json'.format(PYPI_URL, name),
        content_type='application/json',
        body=demo_json,
    )
    assert check(name, u'1.0', session) == (1, assumed_demo_result)



# -*- coding: utf-8 -*-
from pkg_resources import parse_version


def uptodate_analysis(pkginfo, pypiinfo):
    """analyse if used version is current:

    result:
        if empty then most recent
        if 'cfg' in result, some cfg is newer
        if 'pypifinal' in result, some pypi final release is newer
        if 'pypiprerelease' in result, some pypi prerelease is newer

    """
    result = []
    if is_cfg_newer(pkginfo):
        result.append('cfg')
    newer = is_pypi_newer(pypiinfo)
    if newer:
        result.append(newer)
    return result


def is_cfgidx_newer(pkginfo, target_idx):
    """check if a given idx (>0) version is newer than the firstversion

    returns boolean
    """
    vcur = None
    for idx, key in enumerate(pkginfo):
        version = pkginfo[key]['v']
        if not version:
            continue
        if idx == 0:
            vcur = parse_version(version)
        if idx == target_idx:
            return parse_version(version) > vcur
    return False


def is_cfg_newer(pkginfo):
    """ checks if one of the cfg is newer

    returns boolean
    """
    for idx in range(1, len(pkginfo)):
        if is_cfgidx_newer(pkginfo, idx):
            return True


TEST_FINALS = set(['major', 'minor', 'bugfix'])
TEST_PRERELEASE = set(['majorpre', 'minorpre', 'bugfixpre'])


def is_pypi_newer(pypiinfo):
    keys = {_ for _ in pypiinfo if pypiinfo.get(_, False)}
    if TEST_FINALS.intersection(keys):
        return 'pypifinal'
    if TEST_PRERELEASE.intersection(keys):
        return 'pypiprerelease'
    return False



# -*- coding: utf-8 -*-
# inspired partly by dumppickedversions
from pkg_resources import DEVELOP_DIST
from plone.versioncheck.utils import find_relative
from zc.buildout import easy_install

import json
import logging
import os
import sys
import time


logger = easy_install.logger

required_by = {}
versions_by_name = {}

TRACKINGFILENAME = '.plone.versioncheck.tracked.json'


def track_get_dist(old_get_dist):
    def get_dist(self, requirement, *ags, **kw):
        dists = old_get_dist(self, requirement, *ags, **kw)
        for dist in dists:
            dist_name = dist.project_name.lower()
            versions_by_name[dist_name] = (
                dist.version,
                (
                    dist.precedence == DEVELOP_DIST and
                    # heuristics, may fail if not named site-packages
                    'site-packages' not in dist.location and
                    dist.location
                )
            )
            if dist_name not in required_by:
                required_by[dist_name] = []
            if (
                requirement.key != dist_name and
                requirement.key not in required_by[dist_name]
            ):
                required_by[dist_name].append(requirement.key)
            for req in dist.requires():
                if req.key not in required_by:
                    required_by[req.key] = []
                if dist_name not in required_by[req.key]:
                    required_by[req.key].append(dist_name)
        return dists
    return get_dist


def write_tracked(old_logging_shutdown, logfilepath):

    def logging_shutdown():
        # WRITE FILE
        result = {
            'generated': time.time(),
            'required_by': required_by,
            'versions': versions_by_name,
        }
        with open(logfilepath, 'w') as fp:
            json.dump(result, fp, indent=2)
        old_logging_shutdown()
    return logging_shutdown


def install(buildout):
    filepath = os.path.join(
        buildout['buildout']['directory'],
        TRACKINGFILENAME,
    )
    easy_install.Installer.__tracked_versions = {}
    easy_install.Installer._get_dist = track_get_dist(
        easy_install.Installer._get_dist
    )
    logging.shutdown = write_tracked(logging.shutdown, filepath)


def get(pkginfo, buildout):
    filepath = TRACKINGFILENAME
    relative, filename = find_relative(buildout)
    if relative:
        filepath = os.path.join(relative, TRACKINGFILENAME)
    if not os.path.exists(filepath):
        # We are not used as a buildout extension, so this file is
        # not available.
        return
    sys.stderr.write(
        '\nRead tracking information from buildout extension: \n'
        '- {0}\n'.format(
            filepath
        )
    )
    try:
        with open(filepath, 'r') as fp:
            pkginfo['tracking'] = json.load(fp)
    except (IOError, ValueError) as e:
        sys.stderr.write(' - ' + str(e) + '\n')
        return
    delta = time.time() - pkginfo['tracking']['generated']
    days = int(delta // (60 * 60 * 24))
    hours = int(delta // (60 * 60) - days * 60 * 60)
    minutes = int(delta // (60) - days * 60 * 60 - hours * 60)
    seconds = delta % 60
    sys.stderr.write(
        '- age of gathered data: {0:d}d {1:d}h {2:d}m {3:2.3f}s\n'.format(
            days, hours, minutes, seconds,
        )
    )



# -*- coding: utf-8 -*-

from collections import namedtuple
from collections import OrderedDict
from pkg_resources import parse_version
from plone.versioncheck.utils import requests_session

import datetime
import sys


PYPI_URL = 'https://pypi.python.org/pypi'


Release = namedtuple('Release', ['version', 'release_date'])

FLOOR_RELEASE = Release(
    version=u'0.0.0.0',
    release_date=datetime.date(1970, 1, 1),
)


def mmbp_tuple(version):
    """major minor bugfix, postfix tuple from version

    - 1.0     -> 1.0.0.0
    - 1.2     -> 1.2.0.0
    - 1.3.4   -> 1.3.4.0
    - 1.3.4.5 -> 1.3.4.5
    """
    parts = version.base_version.split('.')
    parts += ['0'] * (4 - len(parts))
    return [int(_) for _ in parts]


def check(name, version, session):  # noqa: C901
    result = OrderedDict([
        ('major', FLOOR_RELEASE),
        ('minor', FLOOR_RELEASE),
        ('bugfix', FLOOR_RELEASE),
        ('majorpre', FLOOR_RELEASE),
        ('minorpre', FLOOR_RELEASE),
        ('bugfixpre', FLOOR_RELEASE),
    ])

    # parse version to test against:
    try:
        version = parse_version(version)
    except TypeError:
        return False, 'Version broken/ not checkable.'
    try:
        vtuple = mmbp_tuple(version)
    except ValueError:
        return False, 'Can not check legacy version number.'

    # fetch pkgs json info from pypi
    url = '{url}/{name}/json'.format(url=PYPI_URL, name=name)
    resp = session.get(url)

    # check status code
    if resp.status_code == 404:
        return False, 'Package "{name}" not on pypi.'.format(name=name)
    elif resp.status_code != 200:
        return False, str(resp.status_code)
    data = resp.json()

    # get information about possible updates
    releases = sorted(data['releases'])
    for release in releases:
        # major check (overall)
        rel_v = parse_version(release)
#        if not isinstance(rel_v, SetuptoolsVersion) or not rel_v > version:
        if rel_v <= version:
            continue
        rel_vtuple = mmbp_tuple(rel_v)
        rel_data = data['releases'][release]
        rel_date = datetime.date(1970, 1, 1)
        for rel_pkg in rel_data:
            time_string = rel_pkg.get('upload_time')
            if time_string:
                crel_date = datetime.datetime.strptime(
                    time_string, '%Y-%m-%dT%H:%M:%S').date()
                if crel_date > rel_date:
                    rel_date = crel_date
        if rel_vtuple[0] > vtuple[0]:
            if (
                rel_v.is_prerelease and
                rel_v > parse_version(result['majorpre'].version)
            ):
                result['majorpre'] = Release(version=release,
                                             release_date=rel_date)
            elif (
                not rel_v.is_prerelease and
                rel_v > parse_version(result['major'].version)
            ):
                result['major'] = Release(version=release,
                                          release_date=rel_date)
            continue
        if (  # Only compare same version line
            rel_vtuple[0] == vtuple[0] and
            rel_vtuple[1] > vtuple[1]
        ):
            if (
                rel_v.is_prerelease and
                rel_v > parse_version(result['minorpre'].version)
            ):
                result['minorpre'] = Release(version=release,
                                             release_date=rel_date)
            elif (
                not rel_v.is_prerelease and
                rel_v > parse_version(result['minor'].version)
            ):
                result['minor'] = Release(version=release,
                                          release_date=rel_date)
            continue
        if (  # Only compare same version line
            rel_vtuple[0] == vtuple[0] and
            rel_vtuple[1] == vtuple[1] and
            rel_vtuple[2] > vtuple[2]
        ):
            if (
                rel_v.is_prerelease and
                rel_v > parse_version(result['bugfixpre'].version)
            ):
                result['bugfixpre'] = Release(version=release,
                                              release_date=rel_date)
            elif (
                not rel_v.is_prerelease and
                rel_v > parse_version(result['bugfix'].version)
            ):
                    result['bugfix'] = Release(version=release,
                                               release_date=rel_date)
            continue

    # reset non existing versions
    for version_tag in result.keys():
        if result[version_tag].version == u'0.0.0.0':
            result[version_tag] = None

    # filter out older
    if (
        result['major'] and
        result['majorpre'] and
        parse_version(result['majorpre'].version) <
            parse_version(result['major'].version)
    ):
        result['majorpre'] = None
    if (
        result['minor'] and
        result['minorpre'] and
        parse_version(result['minorpre'].version) <
            parse_version(result['minor'].version)
    ):
        result['minorpre'] = None
    if (
        result['bugfix'] and
        result['bugfixpre'] and
        parse_version(result['bugfixpre'].version) <
            parse_version(result['bugfix'].version)
    ):
        result['bugfixpre'] = None

    return 2 if resp.from_cache else 1, result


def check_all(pkgsinfo, limit=None, nocache=False):
    session = requests_session(nocache=nocache)
    pkgs = pkgsinfo['pkgs']
    sys.stderr.write(
        'Check PyPI for updates of {0:d} packages.'.format(len(pkgs))
    )
    if limit:
        sys.stderr.write(' Check limited to {0:d} packages.'.format(limit))
    pkgsinfo['pypi'] = {}
    errors = []
    for idx, pkgname in enumerate(sorted(pkgs)):
        if not idx % 20 and idx != limit:
            sys.stderr.write('\n{0:4d} '.format(idx))
        current = next(iter(pkgs[pkgname]))
        state, result = check(
            pkgname,
            pkgs[pkgname][current]['v'],
            session
        )
        if not state:
            sys.stderr.write('E')
            errors.append((pkgname, pkgs[pkgname][current]['v'], str(result)))
            continue
        pkgsinfo['pypi'][pkgname] = result
        sys.stderr.write('O' if state == 1 else 'o')
        if limit and idx == limit:
            break
    for error in errors:
        sys.stderr.write(
            '\nError in {0} version {1} reason: {2}'.format(
                *error
            )
        )

    sys.stderr.write('\nPyPI check finished\n')


def update_pkg_info(pkg_name, pkg_data, session):
    for filename, elemdata in pkg_data.items():
        # fetch pkgs json info from pypi
        url = '{url}/{name}/json'.format(url=PYPI_URL, name=pkg_name)
        resp = session.get(url)

        # check status code
        if resp.status_code != 200:
            continue
        data = resp.json()

        rel_date = datetime.date(1970, 1, 1)
        try:
            urls = data['releases'][elemdata['v']]
        except KeyError:
            # release not on PyPI
            continue
        for rel_pkg in urls:
            time_string = rel_pkg.get('upload_time')
            if time_string:
                crel_date = datetime.datetime.strptime(
                    time_string, '%Y-%m-%dT%H:%M:%S').date()
                if crel_date > rel_date:
                    rel_date = crel_date
        elemdata['release_date'] = rel_date

    return True


def update_pkgs_info(pkgsinfo, limit=None, nocache=False):
    session = requests_session(nocache=nocache)
    pkgs = pkgsinfo['pkgs']
    sys.stderr.write(
        'Check PyPI for data of {0:d} packages.'.format(len(pkgs))
    )
    if limit:
        sys.stderr.write(' Check limited to {0:d} packages.'.format(limit))
    errors = []

    idx = 0
    for pkg_name, pkg_data in pkgs.items():
        if not idx % 20 and idx != limit:
            sys.stderr.write('\n{0:4d} '.format(idx))

        state = update_pkg_info(pkg_name, pkg_data, session)
        if not state:
            sys.stderr.write('E')
            errors.append((pkg_name, ))
            continue
        sys.stderr.write('O' if state == 1 else 'o')
        if limit and idx == limit:
            break
        idx += 1

    for error in errors:
        sys.stderr.write(
            '\nError in {0}'.format(
                *error
            )
        )

    sys.stderr.write('\nPyPI check finished\n')


def update_tracking_version_info(pkg_name, pkg_data, session):
    # fetch pkgs json info from pypi
    url = '{url}/{name}/{version}/json'.format(url=PYPI_URL,
                                               name=pkg_name,
                                               # version=pkg_data['version'])
                                               version=pkg_data[0])
    resp = session.get(url)

    # check status code
    if resp.status_code == 404:
        return False, 'Package Version not on pypi.'
    elif resp.status_code != 200:
        return False, str(resp.status_code)
    data = resp.json()

    rel_date = datetime.date(1970, 1, 1)
    for rel_pkg in data['urls']:
        time_string = rel_pkg.get('upload_time')
        if time_string:
            crel_date = datetime.datetime.strptime(
                time_string, '%Y-%m-%dT%H:%M:%S').date()
            if crel_date > rel_date:
                rel_date = crel_date
    # pkg_data['release_date'] = rel_date
    pkg_data.append(rel_date)

    return 2 if resp.from_cache else 1, True


def update_tracking_info(pkgsinfo, nocache=False):
    session = requests_session(nocache=nocache)
    pkgs = pkgsinfo['tracking']['versions']
    sys.stderr.write(
        'Check PyPI for data of {0:d} packages.'.format(len(pkgs))
    )
    errors = []

    idx = 0
    for pkg_name, pkg_data in pkgs.items():
        if not idx % 20:
            sys.stderr.write('\n{0:4d} '.format(idx))

        state, result = update_tracking_version_info(pkg_name,
                                                     pkg_data,
                                                     session)
        if not state:
            sys.stderr.write('E')
            errors.append((pkg_name, str(result)))
            continue
        sys.stderr.write('O' if state == 1 else 'o')
        idx += 1

    for error in errors:
        sys.stderr.write(
            '\nError in {0} reason: {1}'.format(
                *error
            )
        )

    sys.stderr.write('\nPyPI check finished\n')



# -*- coding: utf-8 -*-
from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter
from plone.versioncheck import formatter
from plone.versioncheck import tracking
from plone.versioncheck import utils
from plone.versioncheck.parser import parse
from plone.versioncheck.pypi import check_all
from plone.versioncheck.pypi import update_pkgs_info
from plone.versioncheck.pypi import update_tracking_info

import argparse
import sys


EPILOG = """\
States and color codes:
  [A]ctive (white)
  [D]evelop (green)
  [O]rphaned (magenta)
  [I]nherited (older or same versions are gray, newer are yellow)
  [U]pdate of final release on PyPI available (cyan)
  [P]rerelease update on PyPI available (blue)
  [X] unpinned (red)
  [r] Requirement (gray)
  [a] Annotation (gray)

Color of package name helps to indicate overall state of a package.
"""


parser = ArgumentParser(
    description='Fetch information about pinned versions and its overrides in '
                'simple and complex/cascaded buildouts.',
    epilog=EPILOG,
    formatter_class=RawDescriptionHelpFormatter,
)
parser.add_argument(
    'buildout',
    nargs='?',
    default='buildout.cfg',
    help='path to buildout.cfg or other *.cfg file'
)
parser.add_argument(
    '-p',
    '--pypi',
    help='check PyPI for newer versions',
    action='store_true'
)
parser.add_argument(
    '-n',
    '--newer',
    help='display only packages with newer version than active',
    action='store_true'
)
parser.add_argument(
    '-N',
    '--newer-orphaned',
    help='display orphaned packages only when newer versions available',
    action='store_true'
)
parser.add_argument(
    '-r',
    '--required-by',
    help='show information about requirements (only if tracking file is '
         'available)',
    default=False,
    action='store_true'
)
parser.add_argument(
    '-d',
    '--show-release-dates',
    help='show information about release dates '
         '(only for package lookup from PyPI)',
    default=False,
    action='store_true',
    dest='show_release_dates'
)
parser.add_argument(
    '-i',
    '--ignore-tracking',
    help='ignore tracking file (if present)',
    action='store_true'
)
parser.add_argument(
    '-m',
    '--machine',
    help='show as machine readable output (json)',
    action='store_true'
)
parser.add_argument(
    '--no-cache',
    help='do not use a cache for PyPI',
    action='store_true'
)
parser.add_argument(
    '-b',
    '--browser',
    help='show as html for webbrowser',
    action='store_true'
)
parser.add_argument(
    '-o',
    '--output',
    help='safe output to output-file',
    nargs='?',
    type=argparse.FileType('w'),
    default=sys.stdout
)
parser.add_argument(
    '--no-colors',
    help='do not show colors',
    action='store_true'
)
parser.add_argument(
    '--debug-limit',
    type=int,
    help='Limit the number of PyPI versions fetched for debugging'
)


def run():
    args = parser.parse_args()
    pkgsinfo = {}
    pkgsinfo['pkgs'] = parse(args.buildout)

    # retrieve additional informations
    if not args.ignore_tracking:
        tracking.get(pkgsinfo, args.buildout)
    if args.pypi:
        check_all(pkgsinfo, args.debug_limit, nocache=args.no_cache)
        if args.show_release_dates:
            update_pkgs_info(pkgsinfo, args.debug_limit, nocache=args.no_cache)
            if not args.ignore_tracking:
                update_tracking_info(pkgsinfo, nocache=args.no_cache)

    # Create output
    if args.machine:
        formatter.machine(
            pkgsinfo,
            newer_only=args.newer,
            newer_orphaned_only=args.newer_orphaned,
            limit=args.debug_limit,
            file=args.output,
        )
    elif args.browser:
        formatter.browser(
            pkgsinfo,
            newer_only=args.newer,
            newer_orphaned_only=args.newer_orphaned,
            limit=args.debug_limit,
            show_requiredby=args.required_by,
            show_release_dates=args.show_release_dates,
            file=args.output,
        )
    else:
        utils.COLORED = not args.no_colors
        formatter.human(
            pkgsinfo,
            newer_only=args.newer,
            newer_orphaned_only=args.newer_orphaned,
            limit=args.debug_limit,
            show_requiredby=args.required_by,
        )



# -*- coding: utf-8 -*-

from collections import OrderedDict
from plone.versioncheck.utils import find_relative
from plone.versioncheck.utils import requests_session
from zc.buildout import UserError
from zc.buildout.buildout import Buildout

import contextlib
import os.path
import sys


if sys.version_info < (3, 0):
    from ConfigParser import ConfigParser
    from ConfigParser import NoOptionError
    from ConfigParser import NoSectionError
    from StringIO import StringIO
elif sys.version_info >= (3, 0):
    from configparser import ConfigParser
    from configparser import NoOptionError
    from configparser import NoSectionError
    from io import StringIO


@contextlib.contextmanager
def nostdout():
    save_stdout = sys.stdout
    sys.stdout = StringIO()
    yield
    sys.stdout = save_stdout


def _extract_versions_section(  # NOQA: C901
    session,
    filename,
    base_dir=None,
    version_sections=None,
    annotations=None,
    relative=None,
    version_section_name=None,
    versionannotation_section_name='versionannotations'
):
    if base_dir is None:
        base_dir = os.path.dirname(os.path.abspath(filename))
    if version_sections is None:
        version_sections = OrderedDict()
    if annotations is None:
        annotations = OrderedDict()
    if '://' not in filename:
        if relative and '://' in relative:
            # relative to url!
            filename = '{0}/{1}'.format(relative, filename)
        else:
            if relative:
                if filename.startswith(relative + '/'):
                    filename = filename[len(relative + '/'):]
                filename = os.path.join(base_dir, relative, filename)
            else:
                filename = os.path.join(base_dir, filename)

    sys.stderr.write('\n- {0}'.format(filename))

    try:
        with nostdout():
            buildout = Buildout(filename, [])  # Use zc.buildout parser
    except UserError:
        buildout = {'buildout': {}}
    config = ConfigParser()
    if os.path.isfile(filename):
        config.read(filename)
    elif '://' in filename:
        resp = session.get(filename)
        config.readfp(StringIO(resp.text))
        if resp.from_cache:
            sys.stderr.write('\n  from cache')
        elif resp.status_code != 200:
            sys.stderr.write('\n  ERROR {0:d}'.format(resp.status_code))
        else:
            sys.stderr.write('\n  fresh from server')
    else:
        raise ValueError('{0} does not exist!'.format(filename))

    # first read own versions section
    current_version_section_name = buildout['buildout'].get(
        'versions',
        'versions',
    )
    if version_section_name is None:
        # initial name
        version_section_name = current_version_section_name
    elif version_section_name != current_version_section_name:
        # name changed, not sure if this works as expected! - jensens
        sys.stderr.write(
            '\nName of [versions] (versions = versions) has changed.'
            '\nGlobal versions section name: "{gname}"'
            '\nVersions pinned under that new Section namespace "{nname}"'
            ' will be ignored.'.format(
                gname=version_section_name,
                nname=buildout['buildout'].get('versions')))

    if filename.startswith(base_dir):
        key_name = filename[len(base_dir) + 1:]
    else:
        key_name = filename

    if config.has_section(version_section_name):
        version_sections[key_name] = OrderedDict(
            config.items(version_section_name)
        )
        sys.stderr.write(
            '\n  {0:d} entries in versions section.'.format(
                len(version_sections[key_name])
            )
        )

    # read versionannotations
    versionannotation_section_name = buildout['buildout'].get(
        'versionannotations', versionannotation_section_name)
    if config.has_section(versionannotation_section_name):
        annotations[key_name] = OrderedDict(
            config.items(versionannotation_section_name)
        )
        sys.stderr.write(
            '\n  {0:d} entries in annotations section.'.format(
                len(annotations[key_name])
            )
        )
    try:
        extends = config.get('buildout', 'extends').strip()
    except (NoSectionError, NoOptionError):
        return version_sections, annotations
    for extend in reversed(extends.splitlines()):
        extend = extend.strip()
        if not extend:
            continue
        sub_relative, extend = find_relative(extend, relative)
        _extract_versions_section(
            session,
            extend,
            base_dir,
            version_sections,
            annotations,
            sub_relative,
            version_section_name,
            versionannotation_section_name
        )
    return version_sections, annotations


def parse(buildout_filename, nocache=False):
    sys.stderr.write('Parsing buildout files:')
    if nocache:
        sys.stderr.write('\n(not using caches)')
    base_relative, buildout_filename = find_relative(buildout_filename)
    session = requests_session(nocache=nocache)
    version_sections, annotations = _extract_versions_section(
        session,
        buildout_filename,
        relative=base_relative
    )
    sys.stderr.write('\nparsing finished.\n')
    pkgs = {}

    for name in version_sections:
        for pkg in version_sections[name]:
            if pkg not in pkgs:
                pkgs[pkg] = OrderedDict()

    for pkgname in pkgs:
        pkg = pkgs[pkgname]
        for name in version_sections.keys():
            if pkgname in version_sections.get(name, {}):
                pkg[name] = {'v': version_sections[name][pkgname], 'a': ''}

        for name in annotations.keys():
            if pkgname in annotations.get(name, {}):
                if name in pkg:
                    pkg[name]['a'] = annotations[name][pkgname]
                else:
                    pkg[name] = {'v': None, 'a': annotations[name][pkgname]}

    return pkgs



# -*- coding: utf-8 -*-

from collections import OrderedDict
from plone.versioncheck.formatter import browser
from plone.versioncheck.formatter import machine


pkgsinfo = {'pkgs': {
    'collective.quickupload': OrderedDict([
        ('foo.cfg', {'v': '1.5.8', 'a': ''}),
        ('baz.cfg', {'v': '1.5.2', 'a': ''})
    ]),
    'ipython': OrderedDict([('buildout.cfg', {'v': '5.3.0', 'a': ''})]),
    'lazy': OrderedDict([('buildout.cfg', {'v': '1.0', 'a': ''})]),
    'products.cmfcore': OrderedDict([
        ('buildout.cfg', {'v': '2.1.1', 'a': '\nJust a Test Case\nwith multiple lines'}),  # NOQA: E501
        ('bar.cfg', {'v': '2.2.0', 'a': ''}),
        ('foo.cfg', {'v': '3.0.1', 'a': ''}),
        ('baz.cfg', {'v': '2.2.10', 'a': ''})
    ]),
}}


json_output = '''{
"collective.quickupload": {
    "versions": [
        {
            "description": "foo.cfg",
            "version": "1.5.8",
            "annotation": "",
            "state": "A"
        },
        {
            "description": "baz.cfg",
            "version": "1.5.2",
            "annotation": "",
            "state": "I"
        }
    ],
    "state": "A"
},
"ipython": {
    "versions": [
        {
            "description": "buildout.cfg",
            "version": "5.3.0",
            "annotation": "",
            "state": "A"
        }
    ],
    "state": "A"
},
"lazy": {
    "versions": [
        {
            "description": "buildout.cfg",
            "version": "1.0",
            "annotation": "",
            "state": "A"
        }
    ],
    "state": "A"
},
"products.cmfcore": {
    "versions": [
        {
            "description": "buildout.cfg",
            "version": "2.1.1",
            "annotation": "Just a Test Case\nwith multiple lines",
            "state": "A"
        },
        {
            "description": "bar.cfg",
            "version": "2.2.0",
            "annotation": "",
            "state": "In"
        },
        {
            "description": "foo.cfg",
            "version": "3.0.1",
            "annotation": "",
            "state": "In"
        },
        {
            "description": "baz.cfg",
            "version": "2.2.10",
            "annotation": "",
            "state": "In"
        }
    ],
    "state": "In"
}
}
'''


def test_json_formatter(capsys):
    result = machine(pkgsinfo)
    out, err = capsys.readouterr()
    # Seems that stdout is not correctly captured
    assert result is None
    assert out == ''
    assert err == '\nReport for machines\n\n'
    # assert out == json_output


browser_output = '''<html>
<head>
  <meta charset="utf-8">
  <title>plone.versioncheck</title>
  <style type="text/css" media="screen">
      table {
        font-family: sans-serif;
        font-size: 80%;
        background-color: efefef;
      }
      td {
        border: none;
        padding: 0.5em;
        vertical-align: top;
      }
      thead th {
        border: none;
        padding: 0.5em;
        background-color: #ddddee;
      }
      .even {
        background-color: #efefff;
      }
      .odd {
        background-color: #ffffff;
      }
      .color-D {
        color: green;
      }
      .color-A {
        color: black;
      }
      .color-I {
        color: gray;
      }
      .color-In {
        color: orange;
      }
      .color-U {
        color: darkcyan;
      }
      .color-P {
        color: blue;
      }
      .color-O {
        color: magenta;
      }
      .color-X {
        color: red;
      }

  </style>
</head>
<body>
<table>
  <thead>
    <tr>
      <th>package</th>
      <th>version</th>

      <th>state</th>
      <th>info</th>
      <th>annotation</th>

    </tr>
  </thead>
  <tbody>



    <tr class="odd">


        <td class="color-A" rowspan="2"><span id="collective.quickupload">collective.quickupload</span></td>


        <td class="color-A">1.5.8</td>

        <td class="color-A">A</td>
        <td class="color-A">foo.cfg</td>
        <td class="color-A">&nbsp;</td>

    </tr>

    <tr class="odd">



        <td class="color-I">1.5.2</td>

        <td class="color-I">I</td>
        <td class="color-I">baz.cfg</td>
        <td class="color-I">&nbsp;</td>

    </tr>




    <tr class="even">


        <td class="color-A" rowspan="1"><span id="ipython">ipython</span></td>


        <td class="color-A">5.3.0</td>

        <td class="color-A">A</td>
        <td class="color-A">buildout.cfg</td>
        <td class="color-A">&nbsp;</td>

    </tr>




    <tr class="odd">


        <td class="color-A" rowspan="1"><span id="lazy">lazy</span></td>


        <td class="color-A">1.0</td>

        <td class="color-A">A</td>
        <td class="color-A">buildout.cfg</td>
        <td class="color-A">&nbsp;</td>

    </tr>




    <tr class="even">


        <td class="color-In" rowspan="4"><span id="products.cmfcore">products.cmfcore</span></td>


        <td class="color-A">2.1.1</td>

        <td class="color-A">A</td>
        <td class="color-A">buildout.cfg</td>
        <td class="color-A">Just a Test Case
with multiple lines<br /></td>

    </tr>

    <tr class="even">



        <td class="color-In">2.2.0</td>

        <td class="color-In">I</td>
        <td class="color-In">bar.cfg</td>
        <td class="color-In">&nbsp;</td>

    </tr>

    <tr class="even">



        <td class="color-In">3.0.1</td>

        <td class="color-In">I</td>
        <td class="color-In">foo.cfg</td>
        <td class="color-In">&nbsp;</td>

    </tr>

    <tr class="even">



        <td class="color-In">2.2.10</td>

        <td class="color-In">I</td>
        <td class="color-In">baz.cfg</td>
        <td class="color-In">&nbsp;</td>

    </tr>


  </tbody>
</table>
</body>
</html>'''  # NOQA: E501


def test_browser_formatter(capsys):
    result = browser(pkgsinfo)
    out, err = capsys.readouterr()
    # Seems that stdout is not correctly captured
    assert result is None
    assert out == ''
    assert err == '\nReport for browsers\n\n'
    # assert out == browser_output  # Why is this not working



# -*- coding: utf-8 -*-

from collections import OrderedDict
from plone.versioncheck.parser import nostdout
from plone.versioncheck.parser import parse


def test_nostdout(capsys):
    print('Test stdout')
    out, err = capsys.readouterr()
    assert out == 'Test stdout\n'
    with nostdout():
        print('This should never be in stdout ')

    out, err = capsys.readouterr()
    assert out == ''

    print('This should be printed again.')
    out, err = capsys.readouterr()
    assert out == 'This should be printed again.\n'


def test_parse(capsys):
    input = 'buildout.cfg'

    result = parse(input, False)
    out, err = capsys.readouterr()
    assert err[:23] == "Parsing buildout files:"
    assert """buildout.cfg
  3 entries in versions section.
  1 entries in annotations section.
""" in err

    assert result == {
        'collective.quickupload': OrderedDict([
            ('foo.cfg', {'v': '1.5.8', 'a': ''}),
            ('baz.cfg', {'v': '1.5.2', 'a': ''})
        ]),
        'ipython': OrderedDict([('buildout.cfg', {'v': '5.3.0', 'a': ''})]),
        'lazy': OrderedDict([
            ('buildout.cfg', {'v': '1.0', 'a': ''}),
            ('buildout.d/spam.cfg', {'v': '>= 1.1', 'a': ''})
        ]),
        'products.cmfcore': OrderedDict([
            ('buildout.cfg', {'v': '2.1.1', 'a': '\nJust a Test Case\nwith multiple lines'}),  # NOQA: E501
            ('bar.cfg', {'v': '2.2.0', 'a': ''}),
            ('foo.cfg', {'v': '3.0.1', 'a': ''}),
            ('baz.cfg', {'v': '2.2.10', 'a': ''})
        ]),
    }



# -*- coding: utf-8 -*-

from plone.versioncheck.script import run

import pytest
import sys


help_output = '''
usage: py.test [-h] [-p] [-n] [-N] [-r] [-d] [-i] [-m] [--no-cache] [-b]
               [-o [OUTPUT]] [--no-colors] [--debug-limit DEBUG_LIMIT]
               [buildout]

Fetch information about pinned versions and its overrides in simple and complex/cascaded buildouts.

positional arguments:
  buildout              path to buildout.cfg or other *.cfg file

optional arguments:
  -h, --help            show this help message and exit
  -p, --pypi            check PyPI for newer versions
  -n, --newer           display only packages with newer version than active
  -N, --newer-orphaned  display orphaned packages only when newer versions
                        available
  -r, --required-by     show information about requirements (only if tracking
                        file is available)
  -d, --show-release-dates
                        show information about release dates (only for package
                        lookup from PyPI)
  -i, --ignore-tracking
                        ignore tracking file (if present)
  -m, --machine         show as machine readable output (json)
  --no-cache            do not use a cache for PyPI
  -b, --browser         show as html for webbrowser
  -o [OUTPUT], --output [OUTPUT]
                        safe output to output-file
  --no-colors           do not show colors
  --debug-limit DEBUG_LIMIT
                        Limit the number of PyPI versions fetched for
                        debugging

States and color codes:
  [A]ctive (white)
  [D]evelop (green)
  [O]rphaned (magenta)
  [I]nherited (older or same versions are gray, newer are yellow)
  [U]pdate of final release on PyPI available (cyan)
  [P]rerelease update on PyPI available (blue)
  [X] unpinned (red)
  [r] Requirement (gray)
  [a] Annotation (gray)

Color of package name helps to indicate overall state of a package.
'''  # NOQA: E501


def test_script_help(capsys):
    with pytest.raises(SystemExit):
        sys.argv = ['versioncheck', '--help']
        result = run()
        out, err = capsys.readouterr()
        assert result is None
        assert out == help_output


json_output = '''{
"collective.quickupload": {
    "versions": [
        {
            "description": "foo.cfg",
            "version": "1.5.8",
            "annotation": "",
            "state": "A"
        },
        {
            "description": "baz.cfg",
            "version": "1.5.2",
            "annotation": "",
            "state": "I"
        }
    ],
    "state": "A"
},
"ipython": {
    "versions": [
        {
            "description": "buildout.cfg",
            "version": "5.3.0",
            "annotation": "",
            "state": "A"
        }
    ],
    "state": "A"
},
"lazy": {
    "versions": [
        {
            "description": "buildout.cfg",
            "version": "1.0",
            "annotation": "",
            "state": "A"
        }
    ],
    "state": "A"
},
"products.cmfcore": {
    "versions": [
        {
            "description": "buildout.cfg",
            "version": "2.1.1",
            "annotation": "Just a Test Case\nwith multiple lines",
            "state": "A"
        },
        {
            "description": "bar.cfg",
            "version": "2.2.0",
            "annotation": "",
            "state": "In"
        },
        {
            "description": "foo.cfg",
            "version": "3.0.1",
            "annotation": "",
            "state": "In"
        },
        {
            "description": "baz.cfg",
            "version": "2.2.10",
            "annotation": "",
            "state": "In"
        }
    ],
    "state": "In"
}
}
'''


def test_script_machine(capsys):
    sys.argv = ['versioncheck', '-m']
    result = run()
    out, err = capsys.readouterr()
    assert result is None
    # assert out == json_output


def test_script_browser(capsys):
    sys.argv = ['versioncheck', '-b']
    result = run()
    out, err = capsys.readouterr()
    assert result is None
    # assert out = browser_output


def test_script_pypi(capsys):
    sys.argv = ['versioncheck', '-p']
    result = run()
    out, err = capsys.readouterr()
    assert result is None
    # assert out = browser_output


def test_script_ignore_tracking(capsys):
    sys.argv = ['versioncheck', '-i']
    result = run()
    out, err = capsys.readouterr()
    assert result is None
    # assert out = browser_output


def test_script_ignore_tracking_pypi(capsys):
    sys.argv = ['versioncheck', '-p', '-i']
    result = run()
    out, err = capsys.readouterr()
    assert result is None
    # assert out = browser_output



# -*- coding: utf-8 -*-
from setuptools import find_packages
from setuptools import setup


version = '1.6.9'

long_description = '{0}\n\n{1}'.format(
    open('README.rst').read(),
    open('CHANGES.rst').read()
)

setup(
    name='plone.versioncheck',
    version=version,
    description='Checks pinned versions with overrides in a cascaded buildout',
    long_description=long_description,
    # Get more strings from
    # http://pypi.python.org/pypi?:action=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Buildout',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Quality Assurance',
    ],
    keywords='plone buildout version ',
    author='Jens W. Klein',
    author_email='jens@bluedynamics.com',
    url='https://github.com/plone/plone.versioncheck',
    license='GPL version 2',
    packages=find_packages('src', exclude=['ez_setup']),
    namespace_packages=['plone', ],
    package_dir={'': 'src'},
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'CacheControl[filecache]',
        'colorama',
        'jinja2',
        'requests',
        'setuptools>=12',
        'zc.buildout',
    ],
    setup_requires=[
        'setuptools>=12'
    ],
    extras_require={
        'docs': [
            'Sphinx',
        ],
        'release': [
            'zest.releaser',
        ],
        'test': [
            'pytest',
            'responses',
        ],
        'develop': [
            'pdbpp',
            'isort',
        ],
    },
    entry_points={
        'console_scripts': [
            'versioncheck = plone.versioncheck.script:run',
        ],
        'zc.buildout.extension': [
            'default = plone.versioncheck.tracking:install',
        ]
    },
)



# -*- coding: utf-8 -*-
from __future__ import print_function
from collections import OrderedDict
from jinja2 import Environment
from jinja2 import PackageLoader
from plone.versioncheck import analyser
from plone.versioncheck.utils import color_by_state
from plone.versioncheck.utils import color_dimmed
from plone.versioncheck.utils import color_init
from plone.versioncheck.utils import dots
from plone.versioncheck.utils import get_terminal_size

import datetime
import json
import sys
import textwrap


jenv = Environment(loader=PackageLoader('plone.versioncheck', 'tpl'))


def build_version(
    name,
    pkg,
    pypi,
    tracked,
    key,
    idx,
    flavor='versions',
    orphaned=False,
):
    record = {}
    if flavor == 'versions':
        record['description'] = key
        record['annotation'] = pkg[key]['a'].strip()
        if pkg[key]['v'] is None:
            record['version'] = '(annotation)'
            record['state'] = ''
            return record
        else:
            record['version'] = pkg[key]['v'] or '(unset)'
        if pkg[key].get('release_date', datetime.date(1970, 1, 1)) != datetime.date(1970, 1, 1):  # NOQA: E501
            record['release_date'] = pkg[key].get('release_date')
        if idx == 0:
            if orphaned:
                record['state'] = 'O'
            elif tracked and tracked[1]:
                record['state'] = 'I'
            else:
                record['state'] = 'A'
        elif analyser.is_cfgidx_newer(pkg, idx):
            record['state'] = 'In'
        else:
            record['state'] = 'I'
    else:  # pypi
        record['version'] = pypi[key].version
        record['release_date'] = pypi[key].release_date
        record['description'] = key.capitalize()
        record['annotation'] = None
        if 'pre' in key:
            record['state'] = 'P'
        else:
            record['state'] = 'U'
    return record


def builder(pkgsinfo, newer_only=False, newer_orphaned_only=False, limit=None):  # noqa: C901, E501
    """build
    - OrderedDict with pkgname as keys
    - each entry an record:
      - state: overall package state
      - versions: list of dicts with
        - version number
        - state
        - description
    """
    result = OrderedDict()
    ver_maxlen = 0
    pkgs = pkgsinfo['pkgs']
    pypi = pkgsinfo.get('pypi', {})
    tracked = pkgsinfo.get('tracking', {}).get('versions', {})
    requ = pkgsinfo.get('tracking', {}).get('required_by', {})
    names = sorted(set(tracked.keys()) | set(pkgs.keys()))

    for nidx, name in enumerate(names):
        current_pkg = pkgs.get(name, {})
        record = dict()
        versions = record['versions'] = list()
        unpinned = False
        required_by = requ.get(name, None)
        if required_by:
            record['required_by'] = required_by

        # handle dev-eggs
        devegg = False
        current_tracked = tracked.get(name, None)
        if current_tracked is not None:
            ver_maxlen = max([ver_maxlen, len(current_tracked[0])])
            if current_tracked[1]:
                versions.append({
                    'version': current_tracked[0],
                    'state': 'D',
                    'description': current_tracked[1],
                    'release_date': ''
                })
                devegg = True

        # handle versions.cfg and inherited
        for idx, location in enumerate(current_pkg):
            ver_maxlen = max([ver_maxlen, len(current_pkg[location])])
            versions.append(
                build_version(
                    name,
                    current_pkg,
                    pypi.get(name, {}),
                    current_tracked,
                    location,
                    idx,
                    flavor='versions',
                    orphaned=tracked and current_tracked is None and not devegg
                )
            )
        if not devegg and current_tracked is not None and not len(versions):
            ver_maxlen = max([ver_maxlen, len(current_tracked[0])])
            versions.append({
                'version': current_tracked[0],
                'state': 'X',
                'description': 'unpinned',
                'release_date': current_tracked[2] if len(current_tracked) >= 3 else ''  # NOQA: E501
            })
            unpinned = True

        if pypi.get(name, None) is not None:
            current_pypi = pypi[name]
            for label, version in current_pypi.items():
                if version is None:
                    continue
                ver_maxlen = max([ver_maxlen, len(version)])
                versions.append(
                    build_version(
                        name,
                        current_pkg,
                        current_pypi,
                        current_tracked,
                        label,
                        idx,
                        flavor='pypi',
                    )
                )

        pkgsinfo['ver_maxlen'] = ver_maxlen
        states = analyser.uptodate_analysis(
            current_pkg,
            pypi.get(name, {}),
        )
        if devegg:
            # dev always wins - not true!
            record['state'] = 'D'
        elif unpinned:
            record['state'] = 'X'
        elif tracked and name in pkgs and name not in tracked:
            record['state'] = 'O'
        elif 'pypifinal' in states:
            record['state'] = 'U'
        elif 'cfg' in states:
            record['state'] = 'In'
        elif 'pypifinal' in states:
            record['state'] = 'P'
        else:
            record['state'] = 'A'

        if newer_only and record['state'] == 'A':
            continue
        if (
            newer_orphaned_only and
            record['versions'][0]['state'] == 'O' and
            len(record['versions']) == 1
        ):
            continue

        result[name] = record
    return result


def human(
    pkgsinfo,
    newer_only=False,
    newer_orphaned_only=False,
    limit=None,
    show_requiredby=False
):
    color_init()
    sys.stderr.write('\nReport for humans\n\n')
    data = builder(
        pkgsinfo,
        newer_only=newer_only,
        newer_orphaned_only=newer_orphaned_only,
        limit=limit
    )
    termx, termy = get_terminal_size()
    for name, record in data.items():
        print(color_by_state(record['state']) + name)
        for version in record['versions']:
            print(
                ' ' * 4 +
                color_by_state(version['state']) +
                version['version'] +
                dots(version['version'], pkgsinfo['ver_maxlen']) +
                ' ' + color_by_state(version['state']) +
                version['state'][0] if version['state'] else '' +
                ' ' + version['description']
            )
            if version.get('annotation', None):
                indent = (pkgsinfo['ver_maxlen'] + 5) * ' ' + 'a '
                print(
                    color_dimmed() +
                    textwrap.fill(
                        version['annotation'],
                        termx - pkgsinfo['ver_maxlen'],
                        initial_indent=indent,
                        subsequent_indent=indent,
                    )
                )

        if show_requiredby and record.get('required_by', False):
            req = ' '.join(sorted(record.get('required_by')))
            indent = (pkgsinfo['ver_maxlen'] + 5) * ' ' + 'r '
            print(
                color_dimmed() +
                textwrap.fill(
                    req,
                    termx - pkgsinfo['ver_maxlen'],
                    initial_indent=indent,
                    subsequent_indent=indent,
                )
            )


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, datetime.date):
        serial = obj.isoformat()
        return serial
    raise TypeError('Type not serializable')


def browser(
    pkgsinfo,
    newer_only=False,
    newer_orphaned_only=False,
    limit=None,
    show_requiredby=False,
    show_release_dates=False,
    file=sys.stdout
):
    color_init()
    sys.stderr.write('\nReport for browsers\n\n')
    data = builder(
        pkgsinfo,
        newer_only=newer_only,
        newer_orphaned_only=newer_orphaned_only,
        limit=limit
    )
    template = jenv.get_template('browser.jinja')
    print(template.render(data=data,
                          req_by=show_requiredby,
                          rel_date=show_release_dates,
                          ),
          file=file)


def machine(
    pkgsinfo,
    newer_only=False,
    newer_orphaned_only=False,
    limit=None,
    file=sys.stdout
):
    sys.stderr.write('\nReport for machines\n\n')
    data = builder(
        pkgsinfo,
        newer_only=newer_only,
        newer_orphaned_only=newer_orphaned_only,
        limit=limit
    )
    print(json.dumps(data, indent=4, default=json_serial), file=file)



# -*- coding: utf-8 -*-

from collections import OrderedDict
from plone.versioncheck.analyser import is_cfg_newer
from plone.versioncheck.analyser import is_pypi_newer
from plone.versioncheck.pypi import Release

import datetime


def test_is_cfg_newer():
    pkginfo = {}
    assert is_cfg_newer(pkginfo) is None

    pkginfo = OrderedDict([('foo.cfg', {'a': '',
                                        'release_date': datetime.date(1970, 1, 1),  # NOQA: E501
                                        'v': '1.0.0'}),
                           ('baz.cfg', {'a': '',
                                        'release_date': datetime.date(1970, 1, 1),  # NOQA: E501
                                        'v': '1.0.5'})
                           ]),
    # assert is_cfg_newer(pkginfo) == True


def test_is_pypi_newer():
    pypiinfo = OrderedDict([('major', None),
                            ('minor', None),
                            ('bugfix', None),
                            ('majorpre', None),
                            ('minorpre', None),
                            ('bugfixpre', None)])
    assert is_pypi_newer(pypiinfo) is False

    pypiinfo = OrderedDict([('major', Release(version=u'1.0.0', release_date=datetime.date(1970, 1, 1))),  # NOQA: E501
                            ('minor', None),
                            ('bugfix', None),
                            ('majorpre', None),
                            ('minorpre', None),
                            ('bugfixpre', None)])
    assert is_pypi_newer(pypiinfo) == 'pypifinal'

    pypiinfo = OrderedDict([('major', None),
                            ('minor', Release(version=u'1.1.0', release_date=datetime.date(1970, 1, 1))),  # NOQA: E501
                            ('bugfix', None),
                            ('majorpre', None),
                            ('minorpre', None),
                            ('bugfixpre', None)])
    assert is_pypi_newer(pypiinfo) == 'pypifinal'

    pypiinfo = OrderedDict([('major', None),
                            ('minor', None),
                            ('bugfix', Release(version=u'1.1.1', release_date=datetime.date(1970, 1, 1))),  # NOQA: E501
                            ('majorpre', None),
                            ('minorpre', None),
                            ('bugfixpre', None)])
    assert is_pypi_newer(pypiinfo) == 'pypifinal'

    pypiinfo = OrderedDict([('major', None),
                            ('minor', None),
                            ('bugfix', None),
                            ('majorpre', Release(version=u'2.0.0.a1', release_date=datetime.date(1970, 1, 1))),  # NOQA: E501
                            ('minorpre', None),
                            ('bugfixpre', None)])
    assert is_pypi_newer(pypiinfo) == 'pypiprerelease'

    pypiinfo = OrderedDict([('major', None),
                            ('minor', None),
                            ('bugfix', None),
                            ('majorpre', None),
                            ('minorpre', Release(version=u'1.2.0.a1', release_date=datetime.date(1970, 1, 1))),  # NOQA: E501
                            ('bugfixpre', None)])
    assert is_pypi_newer(pypiinfo) == 'pypiprerelease'

    pypiinfo = OrderedDict([('major', None),
                            ('minor', None),
                            ('bugfix', None),
                            ('majorpre', None),
                            ('minorpre', None),
                            ('bugfixpre', Release(version=u'1.1.2.a1', release_date=datetime.date(1970, 1, 1)))])  # NOQA: E501
    assert is_pypi_newer(pypiinfo) == 'pypiprerelease'
